@@ -84,6 +84,7 @@ struct GdmChooserWidgetPrivate
         guint                    animation_timeout_id;
 
         guint32                  should_hide_inactive_items : 1;
+        guint32                  emit_activated_after_animation : 1;
 
         GdmChooserWidgetPosition separator_position;
         GdmChooserWidgetState    state;
@@ -156,13 +157,86 @@ find_item (GdmChooserWidget *widget,
         return found_item;
 }
 
+typedef struct {
+        GdmChooserWidget           *widget;
+        GdmChooserUpdateForeachFunc func;
+        gpointer                    user_data;
+} UpdateForeachData;
+
+static gboolean
+foreach_item (GtkTreeModel      *model,
+              GtkTreePath       *path,
+              GtkTreeIter       *iter,
+              UpdateForeachData *data)
+{
+        GdkPixbuf *image;
+        char      *name;
+        char      *comment;
+        gboolean   in_use;
+        gboolean   is_separate;
+        gboolean   res;
+        char      *id;
+
+        gtk_tree_model_get (model,
+                            iter,
+                            CHOOSER_ID_COLUMN, &id,
+                            CHOOSER_IMAGE_COLUMN, &image,
+                            CHOOSER_NAME_COLUMN, &name,
+                            CHOOSER_COMMENT_COLUMN, &comment,
+                            CHOOSER_ITEM_IS_IN_USE_COLUMN, &in_use,
+                            CHOOSER_ITEM_IS_SEPARATED_COLUMN, &is_separate,
+                            -1);
+        res = data->func (data->widget,
+                          (const char *)id,
+                          &image,
+                          &name,
+                          &comment,
+                          &in_use,
+                          &is_separate,
+                          data->user_data);
+        if (res) {
+                gtk_list_store_set (GTK_LIST_STORE (model),
+                                    iter,
+                                    CHOOSER_ID_COLUMN, id,
+                                    CHOOSER_IMAGE_COLUMN, image,
+                                    CHOOSER_NAME_COLUMN, name,
+                                    CHOOSER_COMMENT_COLUMN, comment,
+                                    CHOOSER_ITEM_IS_IN_USE_COLUMN, in_use,
+                                    CHOOSER_ITEM_IS_SEPARATED_COLUMN, is_separate,
+                                    -1);
+        }
+
+        g_free (name);
+        g_free (comment);
+        if (image != NULL) {
+                g_object_unref (image);
+        }
+
+        return FALSE;
+}
+
+void
+gdm_chooser_widget_update_foreach_item (GdmChooserWidget           *widget,
+                                        GdmChooserUpdateForeachFunc func,
+                                        gpointer                    user_data)
+{
+        UpdateForeachData fdata;
+
+        fdata.widget = widget;
+        fdata.func = func;
+        fdata.user_data = user_data;
+        gtk_tree_model_foreach (GTK_TREE_MODEL (widget->priv->list_store),
+                                (GtkTreeModelForeachFunc) foreach_item,
+                                &fdata);
+}
+
 static char *
 get_active_item_id (GdmChooserWidget *widget,
                     GtkTreeIter      *iter)
 {
-        char *item_id;
+        char         *item_id;
         GtkTreeModel *model;
-        GtkTreePath *path;
+        GtkTreePath  *path;
 
         g_return_val_if_fail (GDM_IS_CHOOSER_WIDGET (widget), NULL);
 
@@ -310,7 +384,10 @@ shrink_edge_toward_active_row (GdmChooserWidget     *widget,
                 if (relative_position < 0) {
                         gtk_tree_path_next (edge_path);
                 } else {
-                        gtk_tree_path_prev (edge_path);
+                        if (!gtk_tree_path_prev (edge_path)) {
+                                edge_is_hidden = TRUE;
+                                goto out;
+                        }
                 }
                 gtk_tree_row_reference_free (*edge_row);
 
@@ -322,12 +399,15 @@ shrink_edge_toward_active_row (GdmChooserWidget     *widget,
                 gtk_tree_model_filter_convert_iter_to_child_iter (widget->priv->model_filter,
                                                                   &iter, &filtered_iter);
                 gtk_list_store_set (GTK_LIST_STORE (widget->priv->list_store),
-                                    &iter, CHOOSER_ITEM_IS_VISIBLE_COLUMN, FALSE, -1);
+                                    &iter,
+                                    CHOOSER_ITEM_IS_VISIBLE_COLUMN, FALSE,
+                                    -1);
 
                 edge_is_hidden = FALSE;
         } else {
                 edge_is_hidden = TRUE;
         }
+out:
         gtk_tree_path_free (active_path);
         gtk_tree_path_free (edge_path);
 
@@ -335,10 +415,16 @@ shrink_edge_toward_active_row (GdmChooserWidget     *widget,
 }
 
 static gboolean
-iterate_animation (GdmChooserWidget *widget)
+run_animation (GdmChooserWidget *widget,
+               int               number_of_iterations)
 {
         gboolean is_done;
 
+        /* FIXME: this function could be done a lot more efficiently
+         * If we know we need to iterate more than once before the next redraw
+         * we should be able to avoid a lot of intermediate work
+         */
+
         is_done = FALSE;
 
         if (widget->priv->state == GDM_CHOOSER_WIDGET_STATE_SHRINKING) {
@@ -352,7 +438,7 @@ iterate_animation (GdmChooserWidget *widget)
                                 shrink_edge_toward_active_row (widget,
                                                  &widget->priv->bottom_edge_row);
                 }
-        } else {
+        } else if (widget->priv->state == GDM_CHOOSER_WIDGET_STATE_GROWING) {
                 GtkTreePath *path;
                 GtkTreeIter  iter;
                 gboolean     is_visible;
@@ -374,32 +460,81 @@ iterate_animation (GdmChooserWidget *widget)
                                 gtk_tree_path_next (path);
                         } else {
                                 gtk_list_store_set (GTK_LIST_STORE (widget->priv->list_store),
-                                                    &iter, CHOOSER_ITEM_IS_VISIBLE_COLUMN,
-                                                    TRUE, -1);
+                                                    &iter,
+                                                    CHOOSER_ITEM_IS_VISIBLE_COLUMN, TRUE,
+                                                    -1);
                         }
                 } while (is_visible);
 
                 gtk_tree_path_free (path);
+        } else {
+                is_done = TRUE;
         }
 
-        return is_done != TRUE;
+        number_of_iterations--;
+
+        if (number_of_iterations != 0) {
+                is_done = run_animation (widget, number_of_iterations);
+        }
+
+        return is_done;
+}
+
+static gboolean
+on_animation_timeout (GdmChooserWidget *widget)
+{
+        return run_animation (widget, 1) != TRUE;
 }
 
 static void
-stop_animation (GdmChooserWidget *widget)
+on_animation_done (GdmChooserWidget *widget)
 {
         if (widget->priv->animation_timeout_id == 0) {
                 return;
         }
 
-        gtk_tree_row_reference_free (widget->priv->top_edge_row);
-        widget->priv->top_edge_row = NULL;
+        if (widget->priv->top_edge_row != NULL) {
+                gtk_tree_row_reference_free (widget->priv->top_edge_row);
+                widget->priv->top_edge_row = NULL;
+        }
 
-        gtk_tree_row_reference_free (widget->priv->bottom_edge_row);
-        widget->priv->bottom_edge_row = NULL;
+        if (widget->priv->bottom_edge_row != NULL) {
+                gtk_tree_row_reference_free (widget->priv->bottom_edge_row);
+                widget->priv->bottom_edge_row = NULL;
+        }
 
         widget->priv->animation_timeout_id = 0;
         gtk_widget_set_sensitive (GTK_WIDGET (widget), TRUE);
+
+        if (widget->priv->emit_activated_after_animation) {
+                g_signal_emit (widget, signals[ACTIVATED], 0);
+                widget->priv->emit_activated_after_animation = FALSE;
+        }
+}
+
+static int
+get_number_of_on_screen_rows (GdmChooserWidget *widget)
+{
+        GtkTreePath *start_path;
+        GtkTreePath *end_path;
+        int         *start_index;
+        int         *end_index;
+        int          number_of_rows;
+
+        if (!gtk_tree_view_get_visible_range (GTK_TREE_VIEW (widget->priv->items_view),
+                                              &start_path, &end_path)) {
+                return 0;
+        }
+
+        start_index = gtk_tree_path_get_indices (start_path);
+        end_index = gtk_tree_path_get_indices (end_path);
+
+        number_of_rows = *end_index - *start_index;
+
+        gtk_tree_path_free (start_path);
+        gtk_tree_path_free (end_path);
+
+        return number_of_rows;
 }
 
 static void
@@ -417,8 +552,12 @@ start_animation (GdmChooserWidget *widget)
        number_of_rows = gtk_tree_model_iter_n_children (GTK_TREE_MODEL (widget->priv->list_store), NULL);
 
        if (widget->priv->state == GDM_CHOOSER_WIDGET_STATE_SHRINKING) {
+               int number_of_on_screen_rows;
+               int number_of_iterations;
+
                if (number_of_visible_rows <= 1) {
                        widget->priv->state = GDM_CHOOSER_WIDGET_STATE_SHRUNK;
+                       on_animation_done (widget);
                        return;
                }
 
@@ -434,11 +573,18 @@ start_animation (GdmChooserWidget *widget)
                gtk_tree_path_free (edge_path);
 
                g_assert (widget->priv->top_edge_row != NULL && widget->priv->bottom_edge_row != NULL);
+
+               number_of_on_screen_rows = get_number_of_on_screen_rows (widget);
+               number_of_iterations = number_of_visible_rows - number_of_on_screen_rows;
+               number_of_iterations = MAX (0, number_of_iterations);
+
+               run_animation (widget, number_of_iterations);
        }
 
        if (widget->priv->state == GDM_CHOOSER_WIDGET_STATE_GROWING) {
                if (number_of_visible_rows >= number_of_rows) {
                        widget->priv->state = GDM_CHOOSER_WIDGET_STATE_GROWN;
+                       on_animation_done (widget);
                        return;
                }
        }
@@ -451,8 +597,71 @@ start_animation (GdmChooserWidget *widget)
        widget->priv->animation_timeout_id =
                g_timeout_add_full (G_PRIORITY_DEFAULT_IDLE,
                                    1000 / (4 * number_of_rows),
-                                   (GSourceFunc) iterate_animation,
-                                   widget, (GDestroyNotify) stop_animation);
+                                   (GSourceFunc) on_animation_timeout,
+                                   widget, (GDestroyNotify) on_animation_done);
+}
+
+static void
+set_inactive_items_visible (GdmChooserWidget *widget,
+                            gboolean          should_show)
+{
+        GtkTreeModel *model;
+        char         *active_item_id;
+        GtkTreeIter   active_item_iter;
+        GtkTreeIter   iter;
+
+        model = GTK_TREE_MODEL (widget->priv->list_store);
+
+        if (!gtk_tree_model_get_iter_first (model, &iter)) {
+                return;
+        }
+
+        active_item_id = get_active_item_id (widget, &active_item_iter);
+
+        do {
+                gboolean is_active;
+
+                is_active = FALSE;
+                if (active_item_id != NULL) {
+                        char *id;
+
+                        gtk_tree_model_get (model, &iter,
+                                            CHOOSER_ID_COLUMN, &id, -1);
+
+                        if (strcmp (active_item_id, id) == 0) {
+                                is_active = TRUE;
+                                g_free (active_item_id);
+                                active_item_id = NULL;
+                        }
+                        g_free (id);
+                }
+
+                if (!is_active) {
+                        gtk_list_store_set (widget->priv->list_store,
+                                            &iter,
+                                            CHOOSER_ITEM_IS_VISIBLE_COLUMN, should_show,
+                                            -1);
+                } else {
+                        gtk_list_store_set (widget->priv->list_store,
+                                            &iter,
+                                            CHOOSER_ITEM_IS_VISIBLE_COLUMN, TRUE,
+                                            -1);
+                }
+        } while (gtk_tree_model_iter_next (model, &iter));
+
+        g_free (active_item_id);
+}
+
+static void
+skip_animation (GdmChooserWidget *widget)
+{
+        if (widget->priv->state == GDM_CHOOSER_WIDGET_STATE_SHRINKING) {
+                set_inactive_items_visible (GDM_CHOOSER_WIDGET (widget), FALSE);
+                widget->priv->state = GDM_CHOOSER_WIDGET_STATE_SHRUNK;
+        } else if (widget->priv->state == GDM_CHOOSER_WIDGET_STATE_GROWING) {
+                set_inactive_items_visible (GDM_CHOOSER_WIDGET (widget), TRUE);
+                widget->priv->state = GDM_CHOOSER_WIDGET_STATE_GROWN;
+        }
 }
 
 static void
@@ -468,7 +677,12 @@ gdm_chooser_widget_grow (GdmChooserWidget *widget)
         set_frame_text (widget, widget->priv->inactive_text);
 
         widget->priv->state = GDM_CHOOSER_WIDGET_STATE_GROWING;
-        start_animation (widget);
+
+        if (GTK_WIDGET_VISIBLE (widget)) {
+                start_animation (widget);
+        } else {
+                skip_animation (widget);
+        }
 }
 
 static void
@@ -517,8 +731,14 @@ gdm_chooser_widget_shrink (GdmChooserWidget *widget)
                                              GTK_SHADOW_ETCHED_OUT);
 
         clear_selection (widget);
+
         widget->priv->state = GDM_CHOOSER_WIDGET_STATE_SHRINKING;
-        start_animation (widget);
+
+        if (GTK_WIDGET_VISIBLE (widget)) {
+                start_animation (widget);
+        } else {
+                skip_animation (widget);
+        }
 }
 
 static void
@@ -534,11 +754,14 @@ activate_from_row (GdmChooserWidget    *widget,
         }
 
         widget->priv->active_row = gtk_tree_row_reference_copy (row);
-        g_signal_emit (widget, signals[ACTIVATED], 0);
 
         if (widget->priv->should_hide_inactive_items) {
+                widget->priv->emit_activated_after_animation = TRUE;
                 gdm_chooser_widget_shrink (widget);
+        } else {
+                g_signal_emit (widget, signals[ACTIVATED], 0);
         }
+
 }
 
 static void
@@ -551,11 +774,11 @@ deactivate (GdmChooserWidget *widget)
         gtk_tree_row_reference_free (widget->priv->active_row);
         widget->priv->active_row = NULL;
 
-        g_signal_emit (widget, signals[DEACTIVATED], 0, NULL, NULL);
-
         if (widget->priv->state != GDM_CHOOSER_WIDGET_STATE_GROWN) {
                 gdm_chooser_widget_grow (widget);
         }
+
+        g_signal_emit (widget, signals[DEACTIVATED], 0);
 }
 
 static void
@@ -863,6 +1086,20 @@ gdm_chooser_widget_size_allocate (GtkWidget        *widget,
 }
 #endif
 
+static void
+gdm_chooser_widget_hide (GtkWidget *widget)
+{
+        skip_animation (GDM_CHOOSER_WIDGET (widget));
+        GTK_WIDGET_CLASS (gdm_chooser_widget_parent_class)->hide (widget);
+}
+
+static void
+gdm_chooser_widget_show (GtkWidget *widget)
+{
+        skip_animation (GDM_CHOOSER_WIDGET (widget));
+        GTK_WIDGET_CLASS (gdm_chooser_widget_parent_class)->show (widget);
+}
+
 static void
 gdm_chooser_widget_class_init (GdmChooserWidgetClass *klass)
 {
@@ -879,6 +1116,8 @@ gdm_chooser_widget_class_init (GdmChooserWidgetClass *klass)
 #ifdef BUILD_ALLOCATION_HACK
         widget_class->size_allocate = gdm_chooser_widget_size_allocate;
 #endif
+        widget_class->hide = gdm_chooser_widget_hide;
+        widget_class->show = gdm_chooser_widget_show;
 
         signals [ACTIVATED] = g_signal_new ("activated",
                                             G_TYPE_FROM_CLASS (object_class),
@@ -1179,7 +1418,6 @@ add_separator (GdmChooserWidget *widget)
 static gboolean
 update_column_visibility (GdmChooserWidget *widget)
 {
-#if 0
         if (widget->priv->number_of_rows_with_images > 0) {
                 gtk_tree_view_column_set_visible (widget->priv->image_column,
                                                   TRUE);
@@ -1187,7 +1425,6 @@ update_column_visibility (GdmChooserWidget *widget)
                 gtk_tree_view_column_set_visible (widget->priv->image_column,
                                                   FALSE);
         }
-
         if (widget->priv->number_of_in_use_rows > 0) {
                 gtk_tree_view_column_set_visible (widget->priv->is_in_use_column,
                                                   TRUE);
@@ -1195,7 +1432,6 @@ update_column_visibility (GdmChooserWidget *widget)
                 gtk_tree_view_column_set_visible (widget->priv->is_in_use_column,
                                                   FALSE);
         }
-#endif
         return FALSE;
 }
 
@@ -1425,6 +1661,78 @@ gdm_chooser_widget_new (const char *inactive_text,
         return GTK_WIDGET (object);
 }
 
+void
+gdm_chooser_widget_update_item (GdmChooserWidget *widget,
+                                const char       *id,
+                                GdkPixbuf        *new_image,
+                                const char       *new_name,
+                                const char       *new_comment,
+                                gboolean          new_in_use,
+                                gboolean          new_is_separate)
+{
+        GtkTreeModel *model;
+        GtkTreeIter   iter;
+        GdkPixbuf    *image;
+        gboolean      is_separate;
+        gboolean      in_use;
+
+        g_return_if_fail (GDM_IS_CHOOSER_WIDGET (widget));
+
+        model = GTK_TREE_MODEL (widget->priv->list_store);
+
+        if (!find_item (widget, id, &iter)) {
+                g_critical ("Tried to remove non-existing item from chooser");
+                return;
+        }
+
+        is_separate = FALSE;
+        gtk_tree_model_get (model, &iter,
+                            CHOOSER_IMAGE_COLUMN, &image,
+                            CHOOSER_ITEM_IS_IN_USE_COLUMN, &in_use,
+                            CHOOSER_ITEM_IS_SEPARATED_COLUMN, &is_separate,
+                            -1);
+
+        if (image != new_image) {
+                if (image == NULL && new_image != NULL) {
+                        widget->priv->number_of_rows_with_images++;
+                } else if (image != NULL && new_image == NULL) {
+                        widget->priv->number_of_rows_with_images--;
+                }
+                queue_column_visibility_update (widget);
+        }
+        if (image != NULL) {
+                g_object_unref (image);
+        }
+
+        if (in_use != new_in_use) {
+                if (new_in_use) {
+                        widget->priv->number_of_in_use_rows++;
+                } else {
+                        widget->priv->number_of_in_use_rows--;
+                }
+                queue_column_visibility_update (widget);
+        }
+
+        if (is_separate != new_is_separate) {
+                if (new_is_separate) {
+                        widget->priv->number_of_separated_rows++;
+                        widget->priv->number_of_normal_rows--;
+                } else {
+                        widget->priv->number_of_separated_rows--;
+                        widget->priv->number_of_normal_rows++;
+                }
+        }
+
+        gtk_list_store_set (widget->priv->list_store,
+                            &iter,
+                            CHOOSER_IMAGE_COLUMN, new_image,
+                            CHOOSER_NAME_COLUMN, new_name,
+                            CHOOSER_COMMENT_COLUMN, new_comment,
+                            CHOOSER_ITEM_IS_IN_USE_COLUMN, new_in_use,
+                            CHOOSER_ITEM_IS_SEPARATED_COLUMN, new_is_separate,
+                            -1);
+}
+
 void
 gdm_chooser_widget_add_item (GdmChooserWidget *widget,
                              const char       *id,
@@ -1446,10 +1754,12 @@ gdm_chooser_widget_add_item (GdmChooserWidget *widget,
 
         if (in_use) {
                 widget->priv->number_of_in_use_rows++;
+                queue_column_visibility_update (widget);
         }
 
         if (image != NULL) {
                 widget->priv->number_of_rows_with_images++;
+                queue_column_visibility_update (widget);
         }
 
         is_visible = widget->priv->active_row == NULL;
@@ -1501,6 +1811,7 @@ gdm_chooser_widget_remove_item (GdmChooserWidget *widget,
 
         if (is_in_use) {
                 widget->priv->number_of_in_use_rows--;
+                queue_column_visibility_update (widget);
         }
 
         if (is_separate) {
@@ -1556,6 +1867,7 @@ gdm_chooser_widget_set_item_in_use (GdmChooserWidget *widget,
                                     gboolean          is_in_use)
 {
         GtkTreeIter   iter;
+        gboolean      was_in_use;
 
         g_return_if_fail (GDM_IS_CHOOSER_WIDGET (widget));
 
@@ -1563,8 +1875,25 @@ gdm_chooser_widget_set_item_in_use (GdmChooserWidget *widget,
                 return;
         }
 
-        gtk_list_store_set (widget->priv->list_store, &iter,
-                            CHOOSER_ITEM_IS_IN_USE_COLUMN, is_in_use, -1);
+        gtk_tree_model_get (GTK_TREE_MODEL (widget->priv->list_store), &iter,
+                            CHOOSER_ITEM_IS_IN_USE_COLUMN, &was_in_use,
+                            -1);
+
+        if (was_in_use != is_in_use) {
+
+                if (is_in_use) {
+                        widget->priv->number_of_in_use_rows++;
+                } else {
+                        widget->priv->number_of_in_use_rows--;
+                }
+                queue_column_visibility_update (widget);
+
+                gtk_list_store_set (widget->priv->list_store,
+                                    &iter,
+                                    CHOOSER_ITEM_IS_IN_USE_COLUMN, is_in_use,
+                                    -1);
+
+        }
 }
 
 void
