@@ -32,7 +32,6 @@
 #include "gdm-greeter-session.h"
 #include "gdm-greeter-client.h"
 #include "gdm-greeter-panel.h"
-#include "gdm-greeter-background.h"
 #include "gdm-greeter-login-window.h"
 
 #include "gdm-session-manager.h"
@@ -47,7 +46,6 @@ struct GdmGreeterSessionPrivate
 
         GtkWidget             *login_window;
         GtkWidget             *panel;
-        GtkWidget             *background;
 };
 
 enum {
@@ -208,7 +206,6 @@ toggle_panel (GdmSessionManager *manager,
               gboolean           enabled,
               GdmGreeterSession *session)
 {
-#if 0
         if (enabled) {
                 session->priv->panel = gdm_greeter_panel_new ();
                 gtk_widget_show (session->priv->panel);
@@ -216,7 +213,6 @@ toggle_panel (GdmSessionManager *manager,
                 gtk_widget_destroy (session->priv->panel);
                 session->priv->panel = NULL;
         }
-#endif
 }
 
 static void
@@ -335,8 +331,8 @@ launch_metacity (GdmGreeterSession *session)
 static void
 start_window_manager (GdmGreeterSession *session)
 {
-        if (! launch_compiz (session)) {
-                launch_metacity (session);
+        if (! launch_metacity (session)) {
+                launch_compiz (session);
         }
 }
 
@@ -351,9 +347,9 @@ start_settings_daemon (GdmGreeterSession *session)
         ret = FALSE;
 
         error = NULL;
-        g_spawn_command_line_async (LIBEXECDIR "/gdm-settings-daemon", &error);
+        g_spawn_command_line_async (LIBEXECDIR "/gnome-settings-daemon --gconf-prefix=/apps/gdm/simple-greeter/settings-manager-plugins", &error);
         if (error != NULL) {
-                g_warning ("Error starting GDM settings daemon: %s", error->message);
+                g_warning ("Error starting settings daemon: %s", error->message);
                 g_error_free (error);
                 goto out;
         }
@@ -372,12 +368,7 @@ toggle_all_levels (GdmSessionManager *manager,
         if (enabled) {
                 start_settings_daemon (session);
                 start_window_manager (session);
-
-                session->priv->background = gdm_greeter_background_new ();
-                gtk_widget_show (session->priv->background);
         } else {
-                gtk_widget_destroy (session->priv->background);
-                session->priv->background = NULL;
         }
 }
 
